﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Diagnostics;

namespace ConcurrencyAndArrays
{
    class Program
    {
        const int NumThreads = 32;
        const int Range = 10000;
        static int numSamples = 100000000;

        static void Main(string[] args)
        {
            ThreadPool.SetMinThreads(32, 32);

            {
                // Warm up CPU and JIT
                TestHistogram(new Histogram(Range));

                TestHistogram(new HistogramLocked(Range));

                TestHistogram(new HistogramInterlocked(Range));

                TestHistogram(new HistogramThreadBuckets(Range, NumThreads / 8));
                TestHistogram(new HistogramThreadBuckets(Range, NumThreads / 4));
                TestHistogram(new HistogramThreadBuckets(Range, NumThreads / 2));
                TestHistogram(new HistogramThreadBuckets(Range, NumThreads / 1));

                TestHistogram(new HistogramThreadBucketsCacheLine(Range, NumThreads));
            }

            Console.Clear();

            {
                // Real Test
                TestHistogram(new HistogramThreadBucketsCacheLine(Range, NumThreads));

                TestHistogram(new Histogram(Range));

                TestHistogram(new HistogramLocked(Range));

                TestHistogram(new HistogramInterlocked(Range));

                TestHistogram(new HistogramThreadBuckets(Range, NumThreads / 8));
                TestHistogram(new HistogramThreadBuckets(Range, NumThreads / 4));
                TestHistogram(new HistogramThreadBuckets(Range, NumThreads / 2));
                TestHistogram(new HistogramThreadBuckets(Range, NumThreads / 1));

                TestHistogram(new HistogramThreadBucketsCacheLine(Range, NumThreads));
            }
        }

        private static void TestHistogram(Histogram histogram)
        {
            Console.WriteLine("Type: {0}", histogram.GetType().Name);

            int perThreadSamples = numSamples / NumThreads;
            numSamples = perThreadSamples * NumThreads;

            var tasks = new Task[NumThreads];
            var startEvent = new ManualResetEvent(false);

            for (int i = 0; i < NumThreads; ++i)
            {
                tasks[i] = Task.Factory.StartNew((threadId) =>
                {
                    int myThreadId = (int)threadId;
                    var rand = new Random();
                    startEvent.WaitOne();
                    for (int sample = 0; sample < perThreadSamples; ++sample)
                    {
                        histogram.AddSample(rand.Next(0, Range + 1), myThreadId);
                    }
                }, i);
            }

            Console.WriteLine("Adding {0:N0} samples on {1:N0} threads", numSamples, NumThreads);
            var stopwatch = Stopwatch.StartNew();
            startEvent.Set();

            Task.WaitAll(tasks);
            stopwatch.Stop();

            Console.WriteLine("Elapsed: {0}", stopwatch.Elapsed);
            var bucketHistogram = histogram as HistogramThreadBuckets;
            if (bucketHistogram != null)
            {
                Console.WriteLine("ValueCount Buckets: {0}", bucketHistogram.ValueBuckets.Length);
            }

            Console.WriteLine("SampleCount: {0:N0} (Error: {1:N3}%)", histogram.SampleCount, 100.0 * (histogram.SampleCount - numSamples) / (double)numSamples);
            long sum = 0;
            for (int i = 0; i < histogram.SampleValueCounts.Length; ++i) { sum += histogram.SampleValueCounts[i]; }
            Console.WriteLine("Sum: {0:N0} (Error: {1:N3}%)", sum, 100.0 * (sum - numSamples) / (double)numSamples);
            Console.WriteLine();
        }
    }
}
